//  Copyright (C) 2002-2003 Intel Corporation, All Rights Reserved.
//  Permission is hereby granted to merge this program code with 
//  other program material to create a derivative work.  This 
//  derivative work may be distributed in compiled object form only.
//  Any other publication of this program, in any form, without the 
//  explicit permission of the copyright holder is prohibited.
//
//  Send questions and comments to erik.j.johnson@intel.com, 
//  aaron.kunze@intel.com


//-------------------------------------------------------------------
// hello_world.uc - Chapter 4
// Reverses an array of 32-bit quantities in SRAM.
//


// This gives us some basic macros
#include "stdmac.uc"
#include "xbuf.uc"

//-------------------------------------------------------------------
// reverse_array_simple
//
//    Description:
//      This macro reverses an array stored in SRAM and puts 
//		it into another location in SRAM.  This is a simple 
// 		implementation that operates one longword at a time
//
//    Parameters:
//      Outputs: n/a
//      In/Outs: n/a
//      Inputs: in_old - The original array.
//				in_new - The array into which the reversed 
//					     data should go.
//				in_size - The number of 32-bit quantities in 
//						  old.
//      Constants: n/a
//      Labels: n/a
//
//    Side effects: n/a
//
//    See also: n/a
//
//    Example Usage: 
//			reverse_array_simple(old_array, new_array, 10)
//
#macro reverse_array_simple[in_old, in_new, in_size]
.begin
	.reg entries_left current_old_entry 
	.reg current_new_entry 
	.reg $array_data

	// Set up a count of remaining entries and SRAM 
	// pointers
	move(entries_left, in_size)
	// Set up a pointer to the current array entry in 
	// the old array
	move(current_old_entry, in_old)

	// Set up a pointer to the current array entry in 
	// the new array.  This works out to be 
	// in_new + (in_size * 4) - 4, because the code 
	// starts at the back of the new array, and because 
	// it operates with 4-byte longwords.
	move(current_new_entry, in_new)
	add_shf_left(current_new_entry, current_new_entry, 
	             in_size, 2)
	sub(current_new_entry, current_new_entry, 4)

	// Now loop one longword at a time and copy the 
	// array
	.while (entries_left != 0)
		// Need a signal for SRAM accesses
		.sig sram_sig 
		// Read the old array from SRAM
		sram[read, $array_data, current_old_entry, 
			 0, 1], ctx_swap[sram_sig]
		
		// Move the data from the read side of the 
		// transfer register to the write side
		move($array_data, $array_data)

		// Write the new array
		sram[write, $array_data, current_new_entry, 
			 0, 1], ctx_swap[sram_sig]
		
		// Update the counter and the pointers
		sub(entries_left, entries_left, 1)
		add(current_old_entry, current_old_entry, 4)
		sub(current_new_entry, current_new_entry, 4)
	.endw
.end
#endm

//-------------------------------------------------------------------
// reverse_array_advanced
//
//    Description:
// 		This macro reverses an array stored in SRAM and puts 
//		it into another location in SRAM.  This is a more 
//		advanced implementation than the one above because it 
//		operates on 8 longwords at a time.
//
//    Parameters:
//      Outputs: n/a
//      In/Outs: n/a
//      Inputs: in_old - The original array.
//				in_new - The array into which the reversed 
//					     data should go.
//				in_size - The number of 32-bit quantities in 
//						  old.
//      Constants: n/a
//      Labels: n/a
//
//    Side effects: n/a
//
//    See also: n/a
//
//    Example Usage: 
//			reverse_array_advanced(old_array, new_array, 10)
//
#macro reverse_array_advanced[in_old, in_new, in_size]
.begin
	.reg entries_left current_old_entry current_new_entry

	// We need an array of eight transfer registers to 
	// read and write eight longwords at a time
	xbuf_alloc($array_data, 8, read_write)

	// Set up a count of remaining entries and SRAM 
	// pointers
	move(entries_left, in_size)
	// Set up a pointer to the current array entry in 
	// the old array
	move(current_old_entry, in_old)

	// Set up a pointer to the current array entry in 
	// the new array. This works out to be 
	// in_new + (in_size * 4), because the code starts 
	// at the back of the new array, and because it 
	// operates with 4-byte longwords.
	move(current_new_entry, in_new)
	add_shf_left(current_new_entry, current_new_entry, 
		         in_size, 2)

	// Now loop one longword at a time and copy the 
	// array
	.reg entries_to_move   // The number of entries to 
						   // move on each iteration
	.while (entries_left != 0)
		// Need a signal for SRAM accesses
		.sig sram_sig
		
		// Figure out how many entries to move
		.if (entries_left < 8)
			move(entries_to_move, entries_left)
		.else
			immed[entries_to_move, 8]
		.endif

		// Figure out where to write the new data
		sub_shf_left(current_new_entry, 
					 current_new_entry, 
					 entries_to_move, 2)

		// Read the old array from SRAM.  We use an 
		// indirect refernce to specify the number of 
		// entries to read
		.reg sram_indirect
		shf_left(sram_indirect, entries_to_move, 21)
		sub_shf_left(sram_indirect, sram_indirect, 1, 21)
		or_shf_left(sram_indirect, sram_indirect, 1, 25)
		sram[read, $array_data[0], current_old_entry, 0, 
				max_8], ctx_swap[sram_sig], indirect_ref
		
		// When we write the new array, we need to do so 
		// specifying the first transfer register with 
		// the data.  So we need to use indexed transfer 
		// registers.
		.reg xfer_index
		shf_left(xfer_index, &$array_data[0], 2)
		add_shf_left(xfer_index, xfer_index, 
				     entries_to_move, 2)
		sub_shf_left(xfer_index, xfer_index, 1, 2)
		local_csr_wr[T_INDEX, xfer_index]

		// Move the data from the read side of the 
		// transfer registers to the write side
		.reg xfers_to_move

		// Assure the assembler that we know what we're 
		// doing
		.set $array_data1 $array_data2 $array_data3 
		.set $array_data4 $array_data5 $array_data6 
		.set $array_data7

		move(xfers_to_move, entries_to_move)
		// The local_csr_wr above takes 3 cycles to take 
		// effect.  The move above should take care of 
		// one cycle, the nops take care of the other 
		// two
		nop
		nop
		move($array_data[0], *$index--)
		sub(xfers_to_move, xfers_to_move, 1)
		br=0[done_moving_xfers#]
		move($array_data[1], *$index--)
		sub(xfers_to_move, xfers_to_move, 1)
		br=0[done_moving_xfers#]
		move($array_data[2], *$index--)
		sub(xfers_to_move, xfers_to_move, 1)
		br=0[done_moving_xfers#]
		move($array_data[3], *$index--)
		sub(xfers_to_move, xfers_to_move, 1)
		br=0[done_moving_xfers#]
		move($array_data[4], *$index--)
		sub(xfers_to_move, xfers_to_move, 1)
		br=0[done_moving_xfers#]
		move($array_data[5], *$index--)
		sub(xfers_to_move, xfers_to_move, 1)
		br=0[done_moving_xfers#]
		move($array_data[6], *$index--)
		sub(xfers_to_move, xfers_to_move, 1)
		br=0[done_moving_xfers#]
		move($array_data[7], *$index--)

		done_moving_xfers#:

		// Assure the assembler that we know what we're 
		// doing
		.use $array_data0 $array_data1 $array_data2 
		.use $array_data3 $array_data4 $array_data5 
		.use $array_data6 $array_data7

		// Write the new array.  Again we use an 
		// indirect reference, which we have 
		// conveniently saved in a register
		alu[--, --, B, sram_indirect]
		sram[write, $array_data[0], current_new_entry, 
			 0, max_8], ctx_swap[sram_sig], indirect_ref
		
		// Update the counter and the pointers
		sub(entries_left, entries_left, entries_to_move)
		add_shf_left(current_old_entry, 
					 current_old_entry, 
					 entries_to_move, 2)
	.endw

	xbuf_free($array_data)
.end
#endm

// Allocate space for the initial array and the new array
.global_mem old_array     SRAM0 40
.global_mem new_array     SRAM0 40
.global_mem new_array_adv SRAM0 40

// Initialize the arrays
.init old_array        1 2 3 4 5 6 7 8 9 10
.init new_array        0 0 0 0 0 0 0 0 0 0 
.init new_array_adv    0 0 0 0 0 0 0 0 0 0 

// This is the main part of the program
.begin
	// Only do this on one thread
	.if(ctx() == 0)
		ctx_arb[voluntary]
		// Call the reverse_array macro
		start_simple_reverse#:
		reverse_array_simple(old_array, new_array, 10)
		end_simple_reverse#:
		reverse_array_advanced(old_array, new_array_adv, 
							   10)
		end_advanced_reverse#:
	.endif
	ctx_arb[kill]
.end
nop